import filetype
import mimetypes
import os

from decorator import decorator
from pathlib import Path
from os import path
from pyramid.httpexceptions import HTTPNotFound, HTTPFound
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from tutorial_server.deployment import require_tutorial_ready


def resolve_path(base_dir, relative_path, index=('index.html',)):
    """Resolves the ``relative_path`` relative to the ``base_dir``. Will raise
    :class:`~pyramid.httpexceptions.HTTPNotFound` if the ``relative_path`` is not a sub-path of the ``base_dir``
    or no file exists for the resolved path.

    If an ``index`` iterable is provided and the resolved path is a directory, it will test the filenames in
    ``index`` relative to the resolved path. The first matching file is returned.

    :param base_dir:
    :type base_dir: ``str``
    :param relative_path:
    :type relative_path: ``tuple`` of ``str``
    :param index:
    :type index: ``str`` iterable
    :return: The resolved path
    :rtype: :class:`~pathlib.Path`
    """
    file_path = Path(path.join(base_dir, *relative_path))
    file_path = file_path.resolve(strict=True)
    if not str(file_path).startswith(str(Path(base_dir).resolve(strict=True))):
        raise HTTPNotFound()
    if file_path.is_dir():
        matched = None
        for filename in index:
            if file_path.joinpath(filename).is_file():
                matched = filename
                break
        if matched:
            file_path = file_path.joinpath(matched)
    if not file_path.is_file():
        raise HTTPNotFound()
    return file_path


def guess_type(file_path):
    """Guess the mime type of the ``file_path``. Will first attempt to identify the mime type through the file's magic
    number. If that fails, guesses the mime type based on the file extension. Where possible it will also attempt to
    guess the file's encoding.

    :param file_path: The path to get the mime type for
    :type file_path: :class:`~pathlib.Path`
    :return: The mimetype and encoding
    :rtype: ``(mime_type, encoding)``
    """
    file_type = filetype.guess(str(file_path))
    if file_type:
        mime_type = (file_type.mime, None)
    else:
        mime_type = ('text/text', None)
    if mime_type[0] == 'text/text':
        mime_type = mimetypes.guess_type(file_path)
    if not mime_type[0]:
        mime_type = ('text/text', None)
    return mime_type


@view_config(route_name='content.get')
@require_tutorial_ready()
def get_tutorial(request: Request):
    """Fetch a single resource from the tutorial tree."""
    try:
        file_path = resolve_path(os.path.join(request.registry.settings['app.home'], 'tutorial'),
                                 request.matchdict['path'])
        mime_type = guess_type(file_path)
        headerlist = [
            ('X-URL-Prefix', request.registry.settings["url.prefix"][:-len(request.registry.settings["url.app_prefix"])]),
            ('Content-Type', mime_type[0]),
        ]
        if mime_type[1]:
            headerlist = [('Content-Encoding', mime_type[1])]
        return Response(body=file_path.read_bytes(),
                        headerlist=headerlist)
    except FileNotFoundError:
        raise HTTPNotFound()
    except RuntimeError:
        raise HTTPNotFound()



import os
import requests
import shutil
import tarfile

from decorator import decorator
from importlib.resources import read_text
from pyramid.httpexceptions import HTTPClientError
from pyramid.response import Response
from pyramid.view import view_config
from threading import Thread
from zipfile import ZipFile

from tutorial_server import static


tutorial_ready = False


def includeme(config):
    """Setup the deployment handling."""
    config.add_route('ready', f'{config.registry.settings["url.prefix"]}/ready')
    Thread(group=None, target=deploy_tutorial, kwargs={'config': config}).start()


@view_config(route_name='ready', renderer='json')
def ready_view(request):
    """View that returns the ready status."""
    global tutorial_ready

    return {'status': tutorial_ready}


def deploy_tutorial(config):
    """Deploy the tutorial and workspace."""
    global tutorial_ready

    settings = config.registry.settings
    if os.path.exists(settings['app.tmp']):
        shutil.rmtree(settings['app.tmp'])
    deploy_content(settings['app.source'], settings['app.home'], settings['app.tmp'])
    deploy_workspace(settings['app.home'], settings['app.tmp'])
    if os.path.exists(settings['app.tmp']):
        shutil.rmtree(settings['app.tmp'])
    tutorial_ready = True


def deploy_content(app_source, home_dir, tmp_dir):
    """Deploy the tutorial content."""
    if app_source.startswith('http://') or app_source.startswith('https://'):
        source_url, target_filename = app_source.split('$$')
        response = requests.get(source_url)
        if response.status_code == 200:
            os.makedirs(tmp_dir, exist_ok=True)
            with open(os.path.join(tmp_dir, target_filename), 'wb') as out_f:
                out_f.write(response.content)
            deploy_content(os.path.join(tmp_dir, target_filename), home_dir, tmp_dir)
    else:
        if app_source.endswith('.zip'):
            with ZipFile(app_source) as zip_file:
                zip_file.extractall(os.path.join(tmp_dir, 'tutorial'))
            deploy_content(os.path.join(tmp_dir, 'tutorial'), home_dir, tmp_dir)
        elif app_source.endswith('.tar.bz2'):
            with tarfile.open(app_source, mode='r:bz2') as tar_file:
                tar_file.extractall(os.path.join(tmp_dir, 'tutorial'))
            deploy_content(os.path.join(tmp_dir, 'tutorial'), home_dir, tmp_dir)
        elif app_source.endswith('.tar.gz'):
            with tarfile.open(app_source, mode='r:gz') as tar_file:
                tar_file.extractall(os.path.join(tmp_dir, 'tutorial'))
            deploy_content(os.path.join(tmp_dir, 'tutorial'), home_dir, tmp_dir)
        else:
            content_dir = os.path.join(home_dir, 'tutorial')
            if os.path.exists(content_dir):
                shutil.rmtree(content_dir)
            shutil.copytree(app_source, content_dir)


def deploy_workspace(home_dir, tmp_dir):
    """Deploy the tutorial workspace."""
    workspace_source = os.path.join(tmp_dir, 'tutorial', '_static', 'workspace')
    workspace_dir = os.path.join(home_dir, 'workspace')
    if os.path.exists(workspace_source):
        for basepath, _, filenames in os.walk(workspace_source):
            for filename in filenames:
                source_filename = os.path.join(basepath, filename)
                target_filename = os.path.join(workspace_dir, source_filename[len(workspace_source) + 1:])
                if not os.path.exists(target_filename):
                    os.makedirs(os.path.dirname(target_filename), exist_ok=True)
                    shutil.copy2(source_filename, target_filename)


def require_tutorial_ready():
    """Pyramid decorator to check that the tutorial is ready."""
    global tutorial_ready

    def handler(f, *args, **kwargs):
        if tutorial_ready:
            return f(*args, **kwargs)
        else:
            body = read_text(static, 'loading.html').replace('$ready_url', args[0].route_url('ready'))
            return Response(body=body, content_type='text/html', content_encoding='utf-8')
    return decorator(handler)



"""
#####################################################
:mod:`~tutorial_server.views` - Tutorial Server Views
#####################################################

This module contains the view handlers for the Tutorial Server.
"""
import filetype
import mimetypes

from decorator import decorator
from pathlib import Path
from os import path
from pyramid.httpexceptions import HTTPNotFound, HTTPFound
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config


def includeme(config):
    """Setup the Tutorial Server's routes."""
    settings = config.registry.settings

    config.add_route('root', f'{settings["url.prefix"]}/')

    config.add_route('content.get',
                     f'{config.registry.settings["url.prefix"]}/content/*path',
                     request_method='GET')


@view_config(route_name='root')
def root(request: Request):
    raise HTTPFound(request.route_url('content.get', path=()))
